/*
  rlp/component.h

  Copyright (c) 2021 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/
#pragma once

#include "../component.h"
#include "pdu.h"
#include "rlp.h"
#include "sdt/udp.h"

#include <map>
#include <memory>
#include <vector>


namespace ACN::RLP {

/**
 * @brief The RLP::Component class
 *
 * The process, program or application corresponding to a single ACN endpoint.
 * All messages in ACN are sent and received by a component which is identified
 * by a CID.
 */
class Component
  : public ACN::Component
{
public:
  explicit Component(UUID::uuid = UUID::uuid(), std::string fctn = "OpenLCP RLP Component",
                     bool ipv4 = true, bool ipv6 = true);

  /// \brief Set the IPv4 enabled state
  /// \param state
  virtual void setIPv4(const bool state) {enable_IPv4 = state;}
  /// \brief Read the IPv4 enabled state
  /// \return
  bool isEnabledIPv4() const {return enable_IPv4;}

  /// \brief Set the IPv6 enabled state
  /// \param state
  virtual void setIPv6(const bool state) {enable_IPv6 = state;}
  /// \brief Read the IPv6 enabled state
  /// \return
  bool isEnabledIPv6() const {return enable_IPv6;}

  // EPI 17 - ACN RLP on UDP
  virtual void UdpPayloadReceiver(PDU::Stream) const;
  void rlpSendUdp(const uint32_t vector, const PDU::Message<PDU::pdu_data> data,
                  const SDT::UDP::ipAddress&) const;

  // EPI 33 - ACN RLP Operation on TCP
  virtual void TcpPacketReceiver(PDU::Stream) const;
  void rlpSendTcp(const uint32_t vector, const PDU::Message<PDU::pdu_data> data,
                  const SDT::UDP::ipAddress&) const;

protected:
  void RlpReceiver(PDU::Message<RLP::Pdu>) const;
  void RlpRegisterVector(uint32_t, PDU::Handler<RLP::Pdu>);

  virtual void sendUDP(const PDU::Stream, const SDT::UDP::ipAddress&) const;
  virtual void sendTCP(const PDU::Stream, const SDT::UDP::ipAddress&) const;

  bool enable_IPv4;    //!< enable rx/tx with IPv4
  bool enable_IPv6;    //!< enable rx/tx with IPv6

private:
  std::map<uint32_t, std::vector<PDU::Handler<RLP::Pdu>>> rlp_vectors_;
  std::shared_ptr<rlp_header> rlp_header_;
};

}; // ACN::RLP
